/*
  loadchil.c
  Source code for a mex file to read Chilbolton-format data into matlab
  Must be linked against libchil.a

  Copyright (C) 2000 Robin Hogan
*/

#include "mex.h"
#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "chil.h"
#include "mextools.h"
#include "version.h"

#define MEX_PROGRAM_NAME "loadchil" /* I'd use the equiv of argv[0] if it existed */

#define WRITE(a) CHILPrintf(a)
#define JUMPSIZE 512
#define TEMPBUFLEN 20
#define NO_SIG_VAL -999.0
#define PI_OVER_180 0.017453292
#define RADIUS_OF_THE_EARTH 6371.0

unsigned int verbose=0,quiet=0,reqday=0,reqmonth=0,reqyear=0,rangecorrection=0;

int
CreateMatlabArray(char * name,double * data_ptr,int m,int n)
{
  mxArray *field;
  double *pr_ptr;

  if (verbose) CHILPrintf("Writing out matrix: %s\n",name);
  field=mxCreateDoubleMatrix(m,n,mxREAL);
  mxSetName(field,name);
  pr_ptr=mxGetPr(field);
  memcpy(pr_ptr,data_ptr,m*n*sizeof(double));   
  mexPutArray(field,"caller");
  mxFree(data_ptr);
  mxDestroyArray(field); 
  return 0;
}

int
CreateMatlabScalar(char * name,double value)
{
  mxArray *field;

  if (verbose) CHILPrintf("Writing out scalar: %s\n",name);

  if ((field=mxCreateDoubleMatrix(1,1,mxREAL)) == NULL) {
    mexWarnMsgTxt("Unable to create scalar\n");
    return 1;
  }
  mxSetName(field,name);
  *mxGetPr(field)=value;
  mexPutArray(field,"caller");
  mxDestroyArray(field);
  return 0;
}

unsigned int
GetUIntFromMatlabArray(const mxArray * array)
{
  char buf[TEMPBUFLEN];
  unsigned int value;
  if (mxIsChar(array)) {
    mxGetString(array,buf,TEMPBUFLEN);
    sscanf(buf,"%u",&value);
  }
  else {
    value=(unsigned int) *mxGetPr(array);
  }
  return value;
}

float
GetFloatFromMatlabArray(const mxArray * array)
{
  char buf[TEMPBUFLEN];
  float value;
  if (mxIsChar(array)) {
    mxGetString(array,buf,TEMPBUFLEN);
    sscanf(buf,"%f",&value);
  }
  else {
    value=(float) *mxGetPr(array);
  }
  return value;
}

#define MemoryError mexErrMsgTxt("Failed to allocate memory")
void
mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{

  ChilboltonData data;
  ChilboltonFile *chil_file = &(data.chil_file);
  RasterInfo *raster_info = &(data.raster_info);
  Ray *ray = &(data.ray);

  int status,i,j, nfields = 0, iarg = 0;
  int iray=0,iblock=0,nblock=0;
  long int newsize, allocsize;

  double *field_ptr[MAX_FIELDS],*elev_ptr=NULL,*azim_ptr=NULL;
  double *utc_ptr=NULL,* range_ptr=NULL;
  double *date_ptr=NULL,*nez_ptr=NULL;
  double *x_ptr=NULL, *y_ptr=NULL, *z_ptr=NULL, *r_ptr=NULL;
  double nan_value;

  int argc;
  char **argv,*argt;

  char basics = 0, coords = 0;

  /* matlab remembers global variables! must clear them */

  nan_value = mxGetNaN();

  /* Save a lot of time and effort:
     convert all string arguments at the start */
  MexConvertStringArguments(nrhs,prhs,&argc,&argv,&argt, MEX_PROGRAM_NAME);

  CHILSetPrintfFunction(mexPrintf);
  CHILSetCallocFunction(mxCalloc);
  CHILSetFreeFunction(mxFree);
  CHILSetNoSignalValue(NO_SIG_VAL);

  status = CHILInit(argc, argv, &data);

  /* If too few arguments, give a usage description */
  if (nrhs < 2) {
    CHILPrintf("loadchil (" CHIL_VERSION ") -- reads Chilbolton-format radar data into matlab\n"
	       "Contact Robin Hogan <R.J.Hogan@reading.ac.uk> with queries, comments or bugs\n");
    CHILUsage(&data);
    CHILPrintf("\n"
	       "General output options\n"
	       "   -basicsonly    don't write those pointless variables like min_gate...\n"
	       "       -coords    output coordinate matrices x, y, z and r (scanning data only)\n"
	       "  -earthcoords    output coordinate matrices, accounting for earth curvature\n");
   CHILOptions(&data);
   return;
  }
  while (++iarg < argc) {
    if (strcmp(argv[iarg],"-basicsonly") == 0) {
      basics = 1;
    }
    else if (strcmp(argv[iarg],"-coords") == 0) {
      if (!coords) {
	coords = 1;
      }
    }
    else if (strcmp(argv[iarg],"-earthcoords") == 0) {
      coords = 2;
    }
  }
  MexFreeArguments(argv);

#define MINIMUM(a,b) (((a) > (b)) ? (b) : (a))

  if (status == CHIL_OK && CHILFetchRay(&data) == CHIL_OK)
    {
      nfields = MINIMUM(data.maxfields, ray->ntotal);
      if (raster_info->cloud_data) {
	coords = 0;
      }
      allocsize = JUMPSIZE * ray->ngates;
#ifdef DEBUG
      CHILPrintf("Allocating memory for arrays...");
#endif
      if ((range_ptr=(double *) mxCalloc(ray->ngates,sizeof(double))) == NULL) MemoryError;
      if ((elev_ptr=(double *) mxCalloc(JUMPSIZE,sizeof(double))) == NULL) MemoryError;
      if ((azim_ptr=(double *) mxCalloc(JUMPSIZE,sizeof(double))) == NULL) MemoryError;
      if ((utc_ptr=(double *) mxCalloc(JUMPSIZE,sizeof(double))) == NULL) MemoryError;
      if ((date_ptr=(double *) mxCalloc(JUMPSIZE,sizeof(double))) == NULL) MemoryError;
      if (data.process != 0)
	if ((nez_ptr=(double *) mxCalloc(JUMPSIZE,sizeof(double))) == NULL) MemoryError;
      for (i=0;i<nfields;i++)
	if ((field_ptr[i]=(double *) mxCalloc(JUMPSIZE*ray->ngates,sizeof(double))) == NULL) MemoryError;
      if (coords) {
	if ((x_ptr=(double *) mxCalloc(JUMPSIZE*ray->ngates,sizeof(double))) == NULL) MemoryError;
	if ((y_ptr=(double *) mxCalloc(JUMPSIZE*ray->ngates,sizeof(double))) == NULL) MemoryError;
	if ((z_ptr=(double *) mxCalloc(JUMPSIZE*ray->ngates,sizeof(double))) == NULL) MemoryError;
	if ((r_ptr=(double *) mxCalloc(JUMPSIZE*ray->ngates,sizeof(double))) == NULL) MemoryError;
      }
#ifdef DEBUG
      CHILPrintf("done.\n");
#endif
      for (i=0;i<ray->ngates;i++)
	*(range_ptr+i)=(double) ray->range[i]+0.001*ray->range_offset;
      nblock=1;
      iblock=0;
      iray=0;

#ifdef DEBUG 
      fprintf(stderr,"Writing scalars to matlab workspace\n");
#endif

      if (!basics) {
	CreateMatlabScalar("day",(double) ray->day);
	CreateMatlabScalar("month",(double) ray->month);
	CreateMatlabScalar("year",(double) CHILTrueYear(ray->year));
	CreateMatlabScalar("file",(double) raster_info->file);
	CreateMatlabScalar("raster",(double) raster_info->raster);
	CreateMatlabScalar("scan",(double) chil_file->iscan+1);
	CreateMatlabScalar("latitude",(double) CHILLatitude(raster_info));
	CreateMatlabScalar("longitude",(double) CHILLongitude(raster_info));
	CreateMatlabScalar("altitude",(double) CHILAltitude(raster_info));
	CreateMatlabScalar("start_hour",(double) raster_info->hour);
	CreateMatlabScalar("start_min",(double) raster_info->minute);
	CreateMatlabScalar("start_sec",(double) raster_info->second);
	CreateMatlabScalar("min_gate",(double) raster_info->min_gate);
	CreateMatlabScalar("max_gate",(double) raster_info->max_gate);
      }
      CreateMatlabScalar("system",(double) raster_info->system_id);
      CreateMatlabScalar("sweepm",(double) raster_info->sweep_mode);
      CreateMatlabScalar("freq",(double) raster_info->frequency);
      CreateMatlabScalar("drange",(double) ray->avegates * raster_info->gate_spacing / 1000.0);
      CreateMatlabScalar("range_offset",(double) ray->range_offset / 1000.0);

      do
	{
#ifdef DEBUG
	  fprintf(stderr,"%d ",iray);
#endif
	  *(elev_ptr+iray)=(double) ray->elevation;
	  *(azim_ptr+iray)=(double) ray->azimuth;
	  *(utc_ptr+iray)=(double) ray->utc;
	  *(date_ptr+iray)=(double) CHILTrueYear(ray->year)*10000+ray->month*100+ray->day;
	  if (data.process != 0)
	    *(nez_ptr+iray)=(double) ray->nez;

	  if (coords == 1) {
	    for (j=0;j<ray->ngates;j++) {
	      double r = cos(ray->elevation * PI_OVER_180) * range_ptr[j];
	      *(r_ptr+j+iray*ray->ngates) = r;
	      *(x_ptr+j+iray*ray->ngates) = -cos(ray->azimuth * PI_OVER_180) * r;
	      *(y_ptr+j+iray*ray->ngates) = sin(ray->azimuth * PI_OVER_180) * r;
	      *(z_ptr+j+iray*ray->ngates) = sin(ray->elevation * PI_OVER_180) * range_ptr[j];
	    }
	  }
	  else if (coords == 2) {
	    /* Account for curvature of the earth */
	    double D0 = RADIUS_OF_THE_EARTH, D02 = D0 * D0;
	    for (j=0;j<ray->ngates;j++) {
	      double r = cos(ray->elevation * PI_OVER_180) * range_ptr[j];
	      *(r_ptr+j+iray*ray->ngates) = r;
	      *(x_ptr+j+iray*ray->ngates) = -cos(ray->azimuth * PI_OVER_180) * r;
	      *(y_ptr+j+iray*ray->ngates) = sin(ray->azimuth * PI_OVER_180) * r;
	      *(z_ptr+j+iray*ray->ngates) = sin(ray->elevation * PI_OVER_180) * range_ptr[j]
		+ sqrt(D02 + r*r) - D0;
	    }
	  }

#ifdef DEBUG 
	  fprintf(stderr,"%d ",iblock);
#endif

	  /* Store the ray as doubles */
	  for (i = 0; i < nfields; i++) {
	    for (j=0;j<ray->ngates;j++) {
	      if (*(ray->data[i]+j) != NO_SIG_VAL)
		*(field_ptr[i]+j+iray*ray->ngates) = (double) *(ray->data[i]+j);
	      else 
		*(field_ptr[i]+j+iray*ray->ngates) = nan_value;
	    }
	  }
	  iray++;
	  if (++iblock == JUMPSIZE) {
	    if (verbose)
	      CHILPrintf("Increasing storage space for data\n");
	    iblock=0;
	    nblock++;
	    newsize=JUMPSIZE*ray->ngates*nblock;
	    if ((elev_ptr=(double *) mxRealloc(elev_ptr,sizeof(double)*JUMPSIZE*nblock)) == NULL ) MemoryError;
	    if ((azim_ptr=(double *) mxRealloc(azim_ptr,sizeof(double)*JUMPSIZE*nblock)) == NULL ) MemoryError;
	    if ((utc_ptr=(double *) mxRealloc(utc_ptr,sizeof(double)*JUMPSIZE*nblock)) == NULL ) MemoryError; 
	    if ((date_ptr=(double *) mxRealloc(date_ptr,sizeof(double)*JUMPSIZE*nblock)) == NULL ) MemoryError;
	    if (data.process != 0)
	      if ((nez_ptr=(double *) mxRealloc(nez_ptr,sizeof(double)*JUMPSIZE*nblock)) == NULL ) MemoryError;
	    for (i=0;i<nfields;i++) {
#ifdef DEBUG
	      fprintf(stderr,"Re-allocating memory from %ld to %ld doubles\n",allocsize,newsize);
#endif
	      if ((field_ptr[i]=(double *) mxRealloc(field_ptr[i],sizeof(double)*newsize)) == NULL ) MemoryError;
	    }
	    if (coords) {
	      if ((x_ptr=(double *) mxRealloc(x_ptr,sizeof(double)*newsize)) == NULL ) MemoryError;
	      if ((y_ptr=(double *) mxRealloc(y_ptr,sizeof(double)*newsize)) == NULL ) MemoryError;
	      if ((z_ptr=(double *) mxRealloc(z_ptr,sizeof(double)*newsize)) == NULL ) MemoryError;
	      if ((r_ptr=(double *) mxRealloc(r_ptr,sizeof(double)*newsize)) == NULL ) MemoryError;
	    }
	  }
	}
      while ((status = CHILFetchRay(&data)) == CHIL_OK);

      if (iray==0) {
	mexErrMsgTxt("No data on day specified");
      }

      /* resize all our arrays */
      if (verbose) 
	CHILPrintf("Resizing arrays downwards: %d %hu\n",iray,ray->ngates);

      if ((elev_ptr=(double *) mxRealloc(elev_ptr,sizeof(double)*iray)) == NULL ) MemoryError;
      if ((azim_ptr=(double *) mxRealloc(azim_ptr,sizeof(double)*iray)) == NULL ) MemoryError;
      if ((utc_ptr=(double *) mxRealloc(utc_ptr,sizeof(double)*iray)) == NULL ) MemoryError;
      if ((date_ptr=(double *) mxRealloc(date_ptr,sizeof(double)*iray)) == NULL ) MemoryError;
      if (data.process != 0)
	if ((nez_ptr=(double *) mxRealloc(nez_ptr,sizeof(double)*iray)) == NULL ) MemoryError;
      for (i=0;i<nfields;i++) {
	if ((field_ptr[i]=(double *) mxRealloc(field_ptr[i],sizeof(double)*iray*ray->ngates)) == NULL ) MemoryError;
      }
      if (coords) {
	if ((x_ptr=(double *) mxRealloc(x_ptr,sizeof(double)*iray*ray->ngates)) == NULL ) MemoryError;
	if ((y_ptr=(double *) mxRealloc(y_ptr,sizeof(double)*iray*ray->ngates)) == NULL ) MemoryError;
	if ((z_ptr=(double *) mxRealloc(z_ptr,sizeof(double)*iray*ray->ngates)) == NULL ) MemoryError;
	if ((r_ptr=(double *) mxRealloc(r_ptr,sizeof(double)*iray*ray->ngates)) == NULL ) MemoryError;
      }


      if (iray==0) {
	mexWarnMsgTxt("No appropriate data found");
	return;
      }

#ifdef DEBUG 
      fprintf(stderr,"Writing vectors to matlab workspace\n");
#endif
      CreateMatlabArray("range",range_ptr,1,ray->ngates);
      CreateMatlabArray("elev",elev_ptr,1,iray);
      CreateMatlabArray("azim",azim_ptr,1,iray);
      CreateMatlabArray("utc",utc_ptr,1,iray);
      CreateMatlabArray("full_date",date_ptr,1,iray);
      if (data.process != 0)
	CreateMatlabArray("nez",nez_ptr,1,iray);

#ifdef DEBUG 
      fprintf(stderr,"Writing arrays to matlab workspace\n");
#endif
      for (i=0;i<nfields;i++) {
	CreateMatlabArray(CHILFieldName(ray,i,NULL),field_ptr[i],ray->ngates,iray);
      }

      if (coords) {
	CreateMatlabArray("x",x_ptr,ray->ngates,iray);
	CreateMatlabArray("y",y_ptr,ray->ngates,iray);
	CreateMatlabArray("z",z_ptr,ray->ngates,iray);
	CreateMatlabArray("r",r_ptr,ray->ngates,iray);
      }


#ifdef DEBUG 
      fprintf(stderr,"Freeing ray\n");
#endif
      CHILFreeRay(ray);
#ifdef DEBUG 
      fprintf(stderr,"Done\n");
#endif
      if (verbose) 
	CHILPrintf("Done\n");
      CHILEnd(&data);
    }
  else {
    CHILPrintf("Run %s with no options for more information\n",
	       MEX_PROGRAM_NAME);
  }
  return;
}

