#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include <cstdlib>
#include "rplot.h"
#include "chil.h"
#include "mapplot.h"

extern "C" {
#include <netcdf.h>
}

int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    cerr << "NetCDF error on line " << __LINE__ << ": " \
      << nc_strerror(ncstatus) << "\n"; \
    exit(-1); \
} 
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    cerr << "NetCDF error on line " << __LINE__ << ": " \
      << nc_strerror(ncstatus) << " (" << s << ")\n"; \
    exit(-1); \
} 

#define MAX_PLOTS 16
#define NO_SIGNAL -999.0
#define NO_FLOAT_ATTRIBUTE -1.0e30
#define NO_FLOAT_ATTRIBUTE_TEST -0.9e30
#define PI_OVER_180 0.01745329
#define RADIUS_OF_THE_EARTH 6371.0

#define DEFAULT_RAY_DURATION 0.01

//#define DEBUG cerr << "Reached line " << __LINE__ << "\n"
#define DEBUG /* */

#define ERROR(msg,status) cerr << "Error: " << (msg) << "\n"; exit(status)
#define MINIMUM(a,b) (((a) > (b)) ? (b) : (a))
#define MAXIMUM(a,b) (((a) < (b)) ? (b) : (a))

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

static bool verbose = false, quiet = false;

static
char *
request_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  if (nc_inq_atttype(ncid, varid, name, &type) != NC_NOERR) {
    return NULL;
  }
  if (type != NC_CHAR) {
    return NULL;
  }
  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    return NULL;
  }
  value = new char[length+1];
  if (!value) {
    cerr << "Out of memory\n";
    exit(2);
  }
  if (nc_get_att_text(ncid, varid, name, value) != NC_NOERR) {
    return NULL;
  }
  value[length] = '\0';
  return value;
}

static
char *
get_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  report_error(nc_inq_atttype(ncid, varid, name, &type), name);
  if (type != NC_CHAR) {
    cerr << "Error: attribute " << name << " is not of type char\n";
    exit(-1);
  }
  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = new char[length+1];
  if (!value) {
    cerr << "Out of memory\n";
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
short int
get_short_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  short int value;

  report_error(nc_inq_atttype(ncid, varid, name, &type), name);
  if (type != NC_SHORT) {
    cerr << "Error: attribute " << name << " is not of type short int\n";
    exit(-1);
  }
  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  if (length > 1) {
    cerr << "Error: attribute " << name << " has more than 1 element\n";
  }
  report_error(nc_get_att_short(ncid, varid, name, &value), name);
  return value;
}

static
float
get_float_attribute(int ncid, int varid, char *name)
{
  size_t length;
  float value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    return NO_FLOAT_ATTRIBUTE;
  }
  if (length > 1) {
    cerr << "Error: " << name << " attribute has more than 1 element\n";
  }
  report_error(nc_get_att_float(ncid, varid, name, &value), name);
  return value;
}

static
int
get_date(char *datestr, int day, int month, int year)
{
  static char *months[] = {"Jan","Feb","Mar","Apr","May","June",
			   "July","Aug","Sept","Oct","Nov","Dec"};
  if (month < 1 || month > 12) {
    return 1;
  }
  else {
    sprintf(datestr, "%d %s %d", day, months[month-1], year);
  }
  return 0;
}

static
char *
strdup_tolower(const char *in)
{
  char *out = strdup(in);
  char *ch = out;
  if (out) {
    while (*ch) {
      *ch = tolower(*ch);
      ++ch;
    }
  }
  return out;
}

static
void
Usage(char **argv)
{
  cerr << "Usage: " << argv[0] << " [options] file1.nc [file2.nc ...]\n"
    "Options:\n"
    "    -hourrange hr1 hr2  set hour axis range\n"
    "  -heightrange ht1 ht2  set height axis range\n"
    "       -rrange x1 x2    set horizontal axis range for RHIs\n"
    "        -range f1 f2    set colour scale range for all plotted variables\n"
    "       -Zrange Z1 Z2    set colour scale range for Z (dBZ)\n"
    "       -vrange v1 v2    set colour scale range for velocity (m/s)\n"
    "-defaultvrange v1 v2    set velocity range but don't override folding velocity\n"
    "   -widthrange w1 w2    set colour scale range for spectral width (m/s)\n"
    "      -svrange sv1 sv2  set colour scale range for velocity standard deviation (m/s)\n"
    "   -PHIDPrange p1 p2    set the PHIDP colour scale range\n"
    "     -LDRrange l1 l2    set the LDR colour scale range\n"
    "       -nezcal f        calibrate radar by forcing nez to be f dBZ\n"
    "      -betacal f        calibration figure for lidar beta (dB)\n"
    "     -duration hrs      plot only specified number of hours - if negative,\n"
    "                          plot most recent data (overridden by `-hourrange')\n"
    "  -rayduration f        assumed duration of each ray in seconds (default "
       << DEFAULT_RAY_DURATION*3600.0 << ")\n"
    "         -size <x>x<y>  set size of axes in pixels\n"
    "      -barsize <x>x<y>  set size of colourbar in pixels\n"
    "     -fontsize n        set pixel size of font\n"
    "       -system system   specify the name of the instrument\n"
    "        -field field    specify the full name of field being plotted\n"
    "        -units units    specify the units of the field\n"
    "   -fieldtitle          write the name of the field on each panel\n"
    "           -id id       string identifier used in the output filename\n"
    "     -onefield          plot only the first field from each file\n"
    "      -montage          produce only one plot containing all the variables\n"
    "     -onlyplot fields   only plot fields in comma-separated list (e.g. Zh,Zdr)\n"
    "       -stdout          send plot to standard output (implies -montage)\n"
    "          -mag f        set the magnification factor (default 1)\n"
    "      -maxsize d        scale down image if panel is larger than d pixels\n"
    "          -png          set image format to portable network graphics (default)\n"
    "     -ps, -eps          set image format to encapsulated postscript\n"
    "       -prefix dir      write the plots in specified directory\n"
    "      -palette file     read RGB colourmap from a file\n"
    "          -xfs server   specify the X font server (otherwise try unix/:-1,7100,7101,7102)\n" 
    "      -verbose          talk to me\n";
  CHILCreateFileNameUsage();
  exit(0);
}

int
main(int argc, char **argv)
{
  double minr = 0, maxr = 95;

  RFigure *figure[MAX_PLOTS];
  RAxes *axes[MAX_PLOTS];
  RPalette colormap;
  RFontLoader *font_loader;
  ofstream ofs[MAX_PLOTS];

  FILE_FORMAT format=FORMAT_PNG;
  char *xfs = "";
  char titlestr[512], paramstr[512];
  /* these override values are for cloud data only, unless size_override == 1 */
  int width_override = 650, height_override = 230;
  int width = 500, height = 500;
  int bar_width = 16, bar_height = 160;
  int x_offset = 50, y_offset_default = 40;
  int y_offset[MAX_PLOTS];
  float x_mag = 1, y_mag = 1;
  int max_size = 0;
  int fig_width, fig_height;

  int iarg, ifile, ivar;
  int max_vars, nvars;
  int ncid, rangeid, rangedimid, timedimid, timeid, varid[MAX_PLOTS], nezid = 0;
  int elevid, azimid, freqid;
  size_t ngates, nrays, iray;
  float *range, *var, time, nexttime, nez, nez_override = 0.0;
  size_t start[2], count[2];
  int tmp;
  short int year, month, day;

  float minvalue[MAX_PLOTS], maxvalue[MAX_PLOTS], intervalue[MAX_PLOTS];

  float missing_value[MAX_PLOTS], vfold[MAX_PLOTS];
  float minval_override = -15.0, maxval_override = 15.0;
  float minZcloud = -50.0, maxZcloud = 10.0;
  float minZ = -20, maxZ = 40.0;
  float minv = -5.0, maxv = 5.0;
  /* float minwidth = 0.0, maxwidth = 0.8; // This is for a linear scale */
  float minwidth = -1.5, maxwidth = 0; /* This is for a logarithmic scale */
  /* float minsv = 0.0, maxsv = 2.0; // This is for a linear scale */
  float minsv = -3, maxsv = 0; /* This is for a logarithmic scale */
  float betacal = 0.0; /* In bels!!! */
  bool log_scale[MAX_PLOTS], using_nez[MAX_PLOTS];
  char *system = NULL, *system_override = NULL, *system_true = NULL;
  char *long_name = NULL, *long_name_override = NULL;
  char *units = NULL, *units_override = NULL;
  char *suffix, *underscore1 = "", *underscore2 = "";
  char *locfile = NULL;
  char *system_code = "", *prefix = ".", *mapfile = NULL;
  char *filename_format = NULL;
  char *field_list = NULL;
  int font_size = 12;
  float starthour = 0, endhour = 24, minheight = 0, maxheight = 12;
  float duration = 0, dhour, rayduration = DEFAULT_RAY_DURATION;
  float minx = -90, maxx = 90, miny = -90, maxy = 90;
  float angle = 0, dangle = 0, bestangle = 0;
  float minZdr = -1, maxZdr = 4;
  float minLdr = -40, maxLdr = 0;
  float minphidp = -10, maxphidp = 40;

  bool no_bar = false, one_field = false, one_plot = false;
  bool use_duration = false, override_range = false, override_betacal = false;
  bool nez_calibrate = false, setrrange = false, vrange_override = false;
  bool size_override = false, standard_output = false, field_title = false;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-verbose") == 0) {
      verbose = true;
    }
    else if (strcmp(argv[iarg], "-onefield") == 0) {
      one_field = true;
    }
    else if (strcmp(argv[iarg], "-fieldtitle") == 0) {
      field_title = true;
    }
    else if (strcmp(argv[iarg], "-stdout") == 0) {
      one_plot = true;
      standard_output = true;
    }
    else if (strcmp(argv[iarg], "-montage") == 0) {
      one_plot = true;
    }
    else if (strcmp(argv[iarg], "-png") == 0) {
      format = FORMAT_PNG;
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) 
	     || (strcmp(argv[iarg],"-eps") == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    else if (strcmp(argv[iarg],"-hourrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&starthour) < 1) {
	ERROR("Failed to read start hour as float", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&endhour) < 1) {
	ERROR("Failed to read end hour as float", -1);
      }
      use_duration = false;
    }
    else if (strcmp(argv[iarg],"-range") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minval_override) < 1) {
	ERROR("Failed to read colour scale range", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxval_override) < 1) {
	ERROR("Failed to read colour scale range", -1);
      }
      override_range = true;
    }
    else if (strcmp(argv[iarg],"-rrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&minr) < 1) {
	ERROR("Failed to read minimum r as float", -1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%lf",&maxr) < 1) {
	ERROR("Failed to read maximum r as float", -1);
      }
      setrrange=true;
    }
    else if (strcmp(argv[iarg],"-Zrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minZ) < 1) {
	ERROR("Failed to read minimum Z", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxZ) < 1) {
	ERROR("Failed to read maximum Z", -1);
      }
      minZcloud = minZ; maxZcloud = maxZ;
    }
    else if (strcmp(argv[iarg],"-vrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minv) < 1) {
	ERROR("Failed to read minimum velocity", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxv) < 1) {
	ERROR("Failed to read maximum velocity", -1);
      }
      vrange_override = true;
    }
    else if (strcmp(argv[iarg],"-defaultvrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minv) < 1) {
	ERROR("Failed to read minimum velocity", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxv) < 1) {
	ERROR("Failed to read maximum velocity", -1);
      }
    }
    else if (strcmp(argv[iarg],"-widthrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minwidth) < 1) {
	ERROR("Failed to read minimum spectral width", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxwidth) < 1) {
	ERROR("Failed to read maximum spectral width", -1);
      }
    }
    else if (strcmp(argv[iarg],"-svrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minsv) < 1) {
	ERROR("Failed to read minimum velocity standard deviation", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxsv) < 1) {
	ERROR("Failed to read maximum velocity standard deviation", -1);
      }
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&minheight) < 1) {
	ERROR("Failed to read minimum height as float", -1);
      }
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%f",&maxheight) < 1) {
	ERROR("Failed to read maximum height as float", -1);
      }
    }
    else if (strcmp(argv[iarg],"-PHIDPrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minphidp) < 1) {
	ERROR("Failed to read minimum PHIDP as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxphidp) < 1) {
	ERROR("Failed to read maximum PHIDP as float",1);
      }
      if (minphidp >= maxphidp) {
	ERROR("Minimum PHIDP must be less than maximum PHIDP",1);
      }
    }
    else if (strcmp(argv[iarg],"-LDRrange") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&minLdr) < 1) {
	ERROR("Failed to read minimum LDR as float",1);
      }
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&maxLdr) < 1) {
	ERROR("Failed to read maximum LDR as float",1);
      }
      if (minLdr >= maxLdr) {
	ERROR("Minimum LDR must be less than maximum LDR",1);
      }
    }
    else if (strcmp(argv[iarg],"-mag") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%f",&x_mag) < 1) {
	ERROR("Failed to read magnification value",1);
      }
      if (x_mag < 0.01) {
	x_mag = 0.01;
	cerr << "Using magnification factor of " << x_mag << "\n";
      }	
      else if (x_mag > 10) {
	x_mag = 10.0;
	cerr << "Using magnification factor of " << x_mag << "\n";
      }
      y_mag = x_mag;
    }
    else if (strcmp(argv[iarg],"-maxsize") == 0) {
      if (argc == iarg+1 || sscanf(argv[++iarg],"%d",&max_size) < 1) {
	ERROR("Failed to read maximum panel size", 1);
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read font size",1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	ERROR("Font size must be a positive integer",1);
      }
    }
    else if (strcmp(argv[iarg],"-duration") == 0) {
      if (argc > iarg+1) {
	duration = atof(argv[++iarg]);
	if (duration == 0.0) {
	  use_duration = false;
	}
	else {
	  use_duration = true;
	}
      }
      else {
	ERROR("Duration not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-nezcal") == 0) {
      if (argc > iarg+1) {
	nez_override = atof(argv[++iarg]);
	nez_calibrate = true;
      }
      else {
	ERROR("Noise-equivalent reflectivity for calibration not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-betacal") == 0) {
      if (argc > iarg+1) {
	betacal = 0.1 * atof(argv[++iarg]);
	override_betacal = true;
      }
      else {
	ERROR("Beta calibration figure not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-rayduration") == 0) {
      if (argc > iarg+1) {
	rayduration = atof(argv[++iarg])/3600.0;
	if (rayduration <= 0.0) {
	  cerr << "Warning: ray duration less than or equal to 0; using "
	       << DEFAULT_RAY_DURATION*3600.0 << " seconds";
	  rayduration = DEFAULT_RAY_DURATION;
	}
      }
      else {
	ERROR("Ray duration not specified", -1);
      }
    }
    else if (strcmp(argv[iarg], "-size") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &width_override, &height_override) < 2) {
	ERROR("Failed to read size: should be in the form "
	  "<width>x<height>", -1);
      }
      size_override = true;
    }
    else if (strcmp(argv[iarg], "-barsize") == 0) {
      if (argc <= iarg+1 || sscanf(argv[++iarg],"%dx%d",
				   &bar_width,&bar_height) < 2) {
	ERROR("Failed to read bar size: should be in the form "
	      "<width>x<height>", -1);
      }
      if (bar_width < 1 || bar_height < 1) {
	no_bar = true;
      }
    }
    else if (strcmp(argv[iarg], "-fontsize") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read font size", -1);
      }
      font_size = atoi(argv[++iarg]);
      if (font_size < 1) {
	ERROR("Font size must be a positive integer", -1);
      }
    }
    else if (strcmp(argv[iarg], "-units") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read units", -1);
      }
      units_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-system") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read system name", -1);
      }
      system_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-field") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read field name", -1);
      }
      long_name_override = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-onlyplot") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read field names", -1);
      }
      field_list = argv[++iarg];
    }
    else if (strcmp(argv[iarg], "-id") == 0) {
      if (argc <= iarg+1) {
	ERROR("Failed to read system ID", -1);
      }
      system_code = argv[++iarg];
      underscore1 = "_";
    }
    else if (strcmp(argv[iarg],"-prefix") == 0) {
      if (argc > iarg+1) {
	prefix = argv[++iarg];
      }
      else {
	ERROR("Prefix not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-filename") == 0) {
      if (argc > iarg+1) {
	filename_format = argv[++iarg];
      }
      else {
	ERROR("Filename not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-coastline") == 0) {
      if (argc > iarg+1) {
	mapfile = argv[++iarg];
      }
      else {
	ERROR("Coastline file not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-locations") == 0) {
      if (argc > iarg+1) {
	locfile = argv[++iarg];
      }
      else {
	ERROR("Location file not specified",1);
      }
    }
    else if (strcmp(argv[iarg],"-palette") == 0) {
      if (argc > iarg+1) {
	mapfile = argv[++iarg];
      }
      else {
	ERROR("Map file not specified", -1);
      }
    }
    else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	ERROR("X font server not specified", -1);
      }
    }
    else {
      cerr << argv[iarg] << " not understood\n";
      Usage(argv);
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
  }

  if (format == FORMAT_POSTSCRIPT) {
    suffix = "eps";
  }
  else {
    suffix = "png";
  }    

  try {

    /* Contact font server */
    font_loader = new RFontLoader(xfs, format == FORMAT_PNG);
    font_loader->verbose(verbose);

    if (mapfile) {
      if (colormap.load(mapfile)) {
	if (!quiet) {
	  cerr << "Warning: " << mapfile << " not found\n";
	}
	create_jet_palette(&colormap,64);
      }
    }
    else {
      create_jet_palette(&colormap,64);
    }

    /* Loop through all file names on the command line */
    for (ifile = iarg; ifile < argc; ++ifile) {
      size_t ind;
      char *scan_type = "Fixed", *scan_type3, *scan_type3_lower;
      char datestr[128];
      float frequency, start_time, end_time, fixduration;
      float x_scale = 1, y_scale = 1;
      unsigned short int file = 0, raster = 0, scan = 0;
      unsigned short int hour, minute, second;
      int i;
      unsigned char sweep_mode;
      bool cloud_data;

      /* Get dimension sizes */
      handle_error(nc_open(argv[ifile], NC_NOWRITE, &ncid));
      handle_error(nc_inq_dimid(ncid, "time", &timedimid));
      handle_error(nc_inq_dimlen(ncid, timedimid, &nrays));
      handle_error(nc_inq_dimid(ncid, "range", &rangedimid));
      handle_error(nc_inq_dimlen(ncid, rangedimid, &ngates));

      /* Get the ranges */
      handle_error(nc_inq_varid(ncid, "range", &rangeid));
      handle_error(nc_inq_varndims(ncid, rangeid, &tmp));
      if (tmp != 1) {
	ERROR("Range variable is not a 1D vector", -1);
      } 
      range = new float[ngates];
      start[0] = 0;
      count[0] = ngates;
      handle_error(nc_get_vara_float(ncid, rangeid, start, count, range));

      /* Find time variable and get the start and end time */
      handle_error(nc_inq_varid(ncid, "time", &timeid));
      ind = 0;
      handle_error(nc_get_var1_float(ncid, timeid, &ind, &start_time));
      hour = (unsigned short int) start_time;
      minute = (unsigned short int) (60.0 * (start_time - (float) hour));
      second = (unsigned short int) (60.0 * (60.0 * (start_time - (float) hour)
					     - (float) minute));
      ind = nrays-1;
      handle_error(nc_get_var1_float(ncid, timeid, &ind, &end_time));
      fixduration = end_time - start_time;
      if (fixduration <= 0) {
	fixduration += 24.0;
      }
      fixduration = ceil(fixduration * 60.0);

      /* Get the radar frequency */
      ind = 0;
      if (nc_inq_varid(ncid, "frequency", &freqid) == NC_NOERR) {
	handle_error(nc_get_var1_float(ncid, freqid, &ind, &frequency));
      }
      else {
	frequency = 0;
      }

      /* Set the title of the plot (stupidly called "system") and set
	 system_true */
      system_true = get_char_attribute(ncid, NC_GLOBAL, "system");
      if (system_override) {
	system = system_override;
      }
      else {
	system = request_char_attribute(ncid, NC_GLOBAL, "title");
	if (!system) {
	  system = new char[128];
	  sprintf(system, "%d-GHz %s radar", (int) (frequency + 0.49), system_true);
	}
      }

      /* Get the date */
      day = get_short_attribute(ncid, NC_GLOBAL, "day");
      month = get_short_attribute(ncid, NC_GLOBAL, "month");
      year = get_short_attribute(ncid, NC_GLOBAL, "year");
      get_date(datestr, day, month, year);

      /* Is it cloud data? */
      cloud_data = true;
      if (nc_inq_varid(ncid, "azim", &azimid) == NC_NOERR
	  && nc_inq_varid(ncid, "elev", &elevid) == NC_NOERR) {
	int nelevdims = 0, nazimdims = 0;
	if (nc_inq_varndims(ncid, azimid, &nazimdims) == NC_NOERR
	    && nc_inq_varndims(ncid, elevid, &nelevdims) == NC_NOERR
	    && nazimdims == 1 && nelevdims == 1) {
	  cloud_data = false;
	}
      }

      if (!cloud_data) {
	size_t tmp;
	/* Get the scan numbers and the radar frequency */
	/* Accept tape or file attributes */
	if (nc_inq_attlen(ncid, NC_GLOBAL, "tape", &tmp) == NC_NOERR) {
	  file = get_short_attribute(ncid, NC_GLOBAL, "tape");
	}
	else {
	  file = get_short_attribute(ncid, NC_GLOBAL, "file");
	}
	raster = get_short_attribute(ncid, NC_GLOBAL, "raster");
	scan = get_short_attribute(ncid, NC_GLOBAL, "scan");
      }

      /* Get the scan type */
      sweep_mode = UF_VERT;
      scan_type3 = "FIX";
      scan_type3_lower = "fix";
      if (!cloud_data) {
	scan_type = get_char_attribute(ncid, NC_GLOBAL, "scantype");
	if (strcmp(scan_type, "RHI") == 0) {
	  sweep_mode = UF_RHI;
	  scan_type3 = "RHI";
	  scan_type3_lower = "rhi";
	}
	else if (strcmp(scan_type, "Slant-plane") == 0) {
	  sweep_mode = UF_RHI;
	  scan_type3 = "S-P";
	  scan_type3_lower = "s-p";
	}
	else if (strcmp(scan_type, "PPI") == 0) {
	  sweep_mode = UF_PPI;
	  scan_type3 = "PPI";
	  scan_type3_lower = "ppi";
	}
	else if (strcmp(scan_type, "Calibration") == 0) {
	  scan_type3 = "CAL";
	  scan_type3_lower = "cal";
	}
	else if (strcmp(scan_type, "Manual") == 0) {
	  scan_type3 = "MAN";
	  scan_type3_lower = "man";
	}
      }

      if (cloud_data) {
	/*** CLOUD DATA ***/
	width = width_override;
	height = height_override;

	/* Set the hour range if necessary */
	if (use_duration) {
	  start[0] = 0;
	  handle_error(nc_get_var1_float(ncid, timeid, start, &starthour));
	  if (starthour > 0.0 && starthour < duration*0.1) {
	    starthour = 0.0;
	  }
	  if (duration > 0.0) {
	    endhour = starthour + duration;
	  }
	  else {
	    start[0] = nrays-1;
	    handle_error(nc_get_var1_float(ncid, timeid, start, &endhour));
	    if (endhour < 24.0 && endhour-24.0 > duration*0.1) {
	      endhour = 24.0;
	    }
	    if (endhour-starthour < (-duration)) {
	      endhour = starthour - duration;
	    }
	    else if ((starthour = endhour + duration) < 0) {
	      starthour = 0;
	      endhour = -duration;
	    }	   
	  }
	}
	minx = starthour;
	maxx = endhour;
	miny = minheight;
	maxy = maxheight;
      }
      else {
	/*** SCANNING DATA ***/
	size_t index = 0;
	float azim1, azim2;
	float elev1, elev2;
	float ang;

	handle_error(nc_get_var1_float(ncid, azimid, &index, &azim1));
	index = nrays-1;
	handle_error(nc_get_var1_float(ncid, azimid, &index, &azim2));

	index = 0;
	handle_error(nc_get_var1_float(ncid, elevid, &index, &elev1));
	index = nrays-1;
	handle_error(nc_get_var1_float(ncid, elevid, &index, &elev2));

	angle = -1000;

	switch (sweep_mode) {
	case UF_RHI: case UF_COPLANE:
	  sprintf(titlestr,"<B>%s</B>", system);
	  sprintf(paramstr,"%s %02hu:%02hu:%02hu %s %04hu/%03hd/%02hu Azim %5.1f<SUP>o",
		  datestr, hour, minute, second,
		  scan_type3, file, raster, scan, 0.5*(azim1+azim2));

	  y_scale = 4;
	  if (strcmp(system_true, "Singapore") == 0) {
	    x_scale = 4;
	  }
	  miny = minheight;
	  maxy = maxheight;
	  if (setrrange) {
	    minx = minr;
	    maxx = maxr;
	  }
	  else {
	    minx = range[0] * cos(MAXIMUM(elev1, elev2) * PI_OVER_180);
	    maxx = MAXIMUM(range[ngates-1], 90.0/x_scale);
	    if (maxx < minx + 90.0/x_scale) {
	      if (maxx < 90.0/x_scale) {
		minx = 0;
		maxx = 90.0/x_scale;
	      }
	      else {
		minx = maxx - 90.0/x_scale;
	      }
	    }
	  }
	  break;
	case UF_PPI:

	  if (azim1 > azim2) {
	    float tmp_float = azim2;
	    azim2 = azim1;
	    azim1 = tmp_float;	    
	  }

	  minx = miny = 100;
	  maxx = maxy = -100;

	  for (ang = azim1; ang < azim2 + 1.0; ang += 1.0) {
	    /* Note: the +90.0 is because `angle' is CLOCKWISE from DUE WEST */
	    float ang_rad = (ang+90.0) * PI_OVER_180;
	    minx = MINIMUM(minx, range[ngates-1]*(-cos(ang_rad)));
	    minx = MINIMUM(minx, range[0]*(-cos(ang_rad)));
	    maxx = MAXIMUM(maxx, range[ngates-1]*(-cos(ang_rad)));
	    maxx = MAXIMUM(maxx, range[0]*(-cos(ang_rad)));
	    miny = MINIMUM(miny, range[ngates-1]*(sin(ang_rad)));
	    miny = MINIMUM(miny, range[0]*(sin(ang_rad)));
	    maxy = MAXIMUM(maxy, range[ngates-1]*(sin(ang_rad)));
	    maxy = MAXIMUM(maxy, range[0]*(sin(ang_rad)));
	  }
	  if (minx >= maxx-90) {
	    minx -= 0.5 * (90.0-maxx+minx);
	    maxx += 0.5 * (90.0-maxx+minx);
	  }
	  if (miny >= maxy) {
	    miny = -90.0; maxy = 90.0;
	  }
	  sprintf(titlestr,"<B>%s</B>",system);
	  sprintf(paramstr,"%s %02hu:%02hu:%02hu %s %04hu/%03hd/%02hu Elev %4.1f<SUP>o",
		  datestr, hour, minute, second,
		  scan_type3, file, raster, scan, 0.5*(elev1+elev2));
	  break;
	default:
	  y_scale = 4;
	  minx = 0;
	  maxx = fixduration;
	  miny = minheight;
	  maxy = maxheight;
	  x_scale = 100.0/fixduration;
	  sprintf(titlestr,"<B>%s</B>", system);
	  sprintf(paramstr,"%s %02hu:%02hu:%02hu %s %hu/%hd/%hu",
		  datestr, hour, minute, second,
		  scan_type3, file, raster, scan);
	}
	if (size_override) {
	  width = width_override;
	  height = height_override;
	}
	else {
	  width = (int) ((maxx-minx) * x_scale * x_mag * 5);
	  height = (int) ((maxy-miny) * y_scale * y_mag * 5);
	}

	minx = floor(minx);
	miny = floor(miny);
	maxx = ceil(maxx);
	maxy = ceil(maxy);

      }

      /*** ALL DATA ***/
      /* Check width & height do not exceed max_size */
      if (max_size > 0) {
	if (width > height) {
	  if (width > max_size) {
	    int tmp_width = width;
	    width = max_size;
	    height = (max_size * height) / tmp_width;
	  }
	}
	else {
	  if (height > max_size) {
	    int tmp_height = height;
	    height = max_size;
	    width = (max_size * width) / tmp_height;
	  }
	}
      }

      /* Find 2D variables to plot */
      handle_error(nc_inq_nvars(ncid, &nvars));
      max_vars = 0;
      for (ivar = 0; ivar < nvars; ++ivar) {
	char name[NC_MAX_NAME];
	nc_type xtype;
	int ndims, natts;
	int dimids[NC_MAX_VAR_DIMS];

	nc_inq_var(ncid, ivar, name, &xtype, &ndims, dimids, &natts);
   	/* Is it 2D etc? */
	if (ndims == 2 && dimids[0] == timedimid && dimids[1] == rangedimid) {
	  if (field_list) {
	    /* Is it in our list of fields to plot? */
	    char *c1 = field_list;
	    int namelen = strlen(name);
	    bool found = false;

	    while ((c1 = strstr(c1, name))) {
	      char *c2 = c1+namelen;
	      if ((c1 == field_list || c1[-1] == ',')
		  && (*c2 == '\0' || *c2 == ',')) {
		found = true;
		break;
	      }
	      c1 = c2;
	    }
	    if (!found) {
	      /* Skip to next variable */
	      continue;
	    }
	  }
	  varid[max_vars] = ivar;
	  ++max_vars;
	  if (max_vars >= MAX_PLOTS) {
	    break;
	  }
	}
	if (max_vars == 1 && one_field) {
	  break;
	}
      }
      if (!max_vars) {
	/* No fields to plot! */
	delete[] range;
	continue;
      }

      /* Work out the size of the figure */
      if (no_bar) {
	fig_width = width + x_offset + 20;
      }
      else {
	fig_width = width + x_offset + bar_width + 64;
      }
      if (one_plot) {
	fig_height = (height+y_offset_default)*max_vars + 40;
      }
      else {
	fig_height = height + y_offset_default + 40;
      }

      /* Set up each variable for plotting */
      for (i = 0; i < max_vars; i++) {
	RAxes *cbar;
	char tmpstr[256];
	float scale_offset;
	char *field;
	char name[NC_MAX_NAME];
	nc_type xtype;
	int ndims, natts;
	int dimids[NC_MAX_VAR_DIMS];
	int ivar = varid[i];

	nc_inq_var(ncid, ivar, name, &xtype, &ndims, dimids, &natts);

	if (units_override) {
	  units = units_override;
	}
	else {
	  units = request_char_attribute(ncid, ivar, "units_html");
	  if (!units) {
	    units = get_char_attribute(ncid, ivar, "units");
	  }
	}

	if (long_name_override) {
	  long_name = long_name_override;
	}
	else {
	  long_name = get_char_attribute(ncid, ivar, "long_name");
	}

	missing_value[i] = get_float_attribute(ncid, ivar, "missing_value");

	if (one_field || one_plot) {
	  field = "";
	  underscore2 = "";
	}
	else {
	  field = name;
	  underscore2 = "_";
	}

	if (!standard_output) {
	  if (!one_plot || i == 0) {
	    if (filename_format) {
	      /* Use the user-specified format string to create a filename */
	      if (!CHILCreateFileName(tmpstr, filename_format, prefix,
				      file, raster, scan,
				      year, month, day,
				      hour, minute, second,
				      system_true, scan_type)) {
		cerr << "Error interpretting filename format \""
		     << filename_format << "\": "
		     << tmpstr << "\n";
		break;
	      }
	    }
	    else if (cloud_data) {
	      sprintf(tmpstr, "%s/%04hd%02hd%02hd%s%s%s%s.%s",
		      prefix, year, month, day, underscore1,
		      system_code, underscore2, field, suffix);
	    }
	    else {
	      //	      char *system_lower = strdup_tolower(system_true);
	      sprintf(tmpstr, "%s/%04hdr%03hds%02hd_%02hd%02hd%02hd_%s_%s%s%s%s%s.%s",
		      prefix, file, raster, scan, hour, minute, second,
		      scan_type3, system_true, underscore1, system_code,
		      underscore2, field, suffix);
	    }
	    ofs[i].open(tmpstr);

	    if (!ofs[i]) {
	      cerr << "Failed to write " << tmpstr << "\n";
	      exit(-1);
	    }

	    if (!quiet) {
	      if (!i) {
		cerr << "Writing";
	      }
	      cerr << " " << tmpstr ;
	    }

	    /* Initialise plot */
	    if (format == FORMAT_POSTSCRIPT) {
	      figure[i] = new RFigure(fig_width, fig_height, POSTSCRIPT,
				      ofs+i, system, 0.5);
	    }
	    else {
	      figure[i] = new RFigure(fig_width, fig_height, BITMAP_8BIT);
	    }
	  }
	  else {
	    /* Use the same figure */
	    figure[i] = figure[0];
	  }
	}
	else if (!one_plot || i == 0) /* One plot to standard output */ {
	  /* Initialise plot */
	  if (format == FORMAT_POSTSCRIPT) {
	    figure[i] = new RFigure(fig_width, fig_height, POSTSCRIPT,
				    &cout, system, 0.5);
	  }
	  else {
	    figure[i] = new RFigure(fig_width, fig_height, BITMAP_8BIT);
	  }
	}
	else {
	  /* Use the same figure */
	  figure[i] = figure[0];
	}

	if (!one_plot || i == 0) {
	  figure[i]->clear();
	  figure[i]->axis(0,0,fig_width, fig_height);
	  if (cloud_data) {
	    figure[i]->font = font_loader->load(font_size,
						string("helvetica"),
						string("bold"),
						string("r"),
						string("adobe"));
	    figure[i]->text(0.01, 0.0, system, ANCHOR_BOTTOMLEFT);
	  }
	  figure[i]->font = font_loader->load(font_size,
					      string("helvetica"),
					      string("medium"),
					      string("r"),
					      string("adobe"));
	}

	y_offset[i] = y_offset_default+(height+y_offset_default)*one_plot*(max_vars-i-1);
	axes[i] = new RAxes(figure[i], x_offset, y_offset[i],
			    x_offset + width, y_offset[i] + height);

	axes[i]->axis(minx, miny, maxx, maxy);
	if (!one_plot || i == 0) {
	  if (cloud_data) {
	    get_date(tmpstr, day, month, year);
	    axes[i]->text(minx, maxy, tmpstr, ANCHOR_BOTTOMLEFT);
	  }
	  else {
	    axes[i]->text(0.5*(minx+maxx), maxy + ((float) (16*(maxy-miny)))/height,
			  titlestr, ANCHOR_BOTTOM);
	    axes[i]->text(0.5*(minx+maxx), maxy, paramstr, ANCHOR_BOTTOM);
	  }
	}
	axes[i]->clipbox_to_axis();
  
	log_scale[i] = false;
	using_nez[i] = false;
	vfold[i] = NO_FLOAT_ATTRIBUTE;
	if (cloud_data) {
	  minvalue[i] = minZcloud;
	  maxvalue[i] = maxZcloud;
	}
	else {
	  minvalue[i] = minZ;
	  maxvalue[i] = maxZ;
	}
	intervalue[i] = 10;
	scale_offset = 0;

	if (strncmp(name, "beta", 4) == 0) {
	  float scale_factor;

	  scale_offset = -7;
	  log_scale[i] = true;
	  minvalue[i] = 0; maxvalue[i] = 3; intervalue[i] = 1;

	  scale_factor = get_float_attribute(ncid, ivar, "scale_factor");
	  if (scale_factor >= NO_FLOAT_ATTRIBUTE_TEST && !override_betacal) {
	    betacal = log10(scale_factor) - scale_offset;
	  }
	}
	else if (name[0] == 'v') {
	  vfold[i] = get_float_attribute(ncid, ivar, "folding_velocity");
	  if (vrange_override || vfold[i] < NO_FLOAT_ATTRIBUTE_TEST) {
	    minvalue[i] = minv; maxvalue[i] = maxv;
	  }
	  else {
	    minvalue[i] = -vfold[i]; maxvalue[i] = +vfold[i];
	  }
	  if (maxvalue[i]-minvalue[i] < 6) {
	    intervalue[i] = 1;
	  }
	  else if (maxvalue[i]-minvalue[i] < 9) {
	    intervalue[i] = 2;
	  }
	  else if (maxvalue[i]-minvalue[i] < 11) {
	    intervalue[i] = 2.5;
	  }
	  else if (maxvalue[i]-minvalue[i] < 22) {
	    intervalue[i] = 5; 
	  }
	  else {
	    intervalue[i] = 10;
	  }
	}
	else if (strncmp(name, "sigma_v", 7) == 0) {
	  log_scale[i] = true;
	  minvalue[i] = minsv; maxvalue[i] = maxsv; intervalue[i] = 1;
	}
	else if (strncmp(name, "width", 5) == 0) {
	  log_scale[i] = true;
	  minvalue[i] = minwidth; maxvalue[i] = maxwidth; intervalue[i] = 0.5;
	}
	else if (nez_calibrate && strncmp(name, "Zh", 2) == 0) {
	  if (nc_inq_varid(ncid, "nez", &nezid) == NC_NOERR) {
	    using_nez[i] = true;
	  }
	}
	else if (strncmp(name, "Zdr", 3) == 0) {
	  minvalue[i] = minZdr; maxvalue[i] = maxZdr; intervalue[i] = 1;
	}
	else if (strncmp(name, "Ldr", 3) == 0) {
	  minvalue[i] = minLdr; maxvalue[i] = maxLdr; intervalue[i] = 10;
	}
	else if (strncmp(name, "phi_dp", 6) == 0) {
	  minvalue[i] = minphidp; maxvalue[i] = maxphidp; intervalue[i] = 10;
	}

	if (override_range) {
	  minvalue[i] = minval_override - scale_offset;
	  maxvalue[i] = maxval_override - scale_offset;
	}

	if (!no_bar) {
	  cbar = new RAxes(figure[i], (fig_width-10)-bar_width,
			   y_offset[i]+0.5*(height-bar_height),
			   fig_width-10, y_offset[i]+0.5*(height+bar_height));
	  cbar->colorbar(&colormap,VERTICAL);
	  cbar->box();
	  cbar->axis(0,minvalue[i]+scale_offset,1,maxvalue[i]+scale_offset);
	  if (log_scale[i]) {
	    cbar->yticks(minvalue[i]+scale_offset,intervalue[i],"10<SUP>%g",0);
	  }
	  else {
	    cbar->yticks(minvalue[i]+scale_offset,intervalue[i],"%g",0);
	  }
       	  if (!field_title) {
	    sprintf(tmpstr,"%s (%s)", long_name, units);
	    cbar->ylabel(tmpstr);
	  }
	  else {
	    cbar->ylabel(units);
	  }
	  delete cbar;
	}

	/* Clean up */
	if (units && !units_override) {
	  delete[] units;
	  units = NULL;
	}
	if (long_name && !long_name_override) {
	  delete[] long_name;
	  long_name = NULL;
	}
      }

      if (!quiet && !standard_output) {
	cerr << "\n";
      }

      var = new float[ngates];
      start[1] = 0;
      count[0] = 1;
      count[1] = ngates;

      for (iray = 0; iray < nrays; ++iray) {
	float elev, azim;

	start[0] = iray;
	handle_error(nc_get_var1_float(ncid, timeid, start, &time));
	if (!cloud_data) {
	  handle_error(nc_get_var1_float(ncid, azimid, start, &azim));
	  handle_error(nc_get_var1_float(ncid, elevid, start, &elev));
	}
	if (iray < nrays-1) {
	  start[0] = iray+1;
	  handle_error(nc_get_var1_float(ncid, timeid, start, &nexttime));
	  if (nexttime - time > rayduration || nexttime < time) {
	    nexttime = time + rayduration;
	  }
	  start[0] = iray;
	}
	else {
	  nexttime = time + rayduration;
	}

	if (cloud_data) {
	  /*** CLOUD DATA ***/
	  for (ivar = 0; ivar < max_vars; ++ivar) {
	    float gap = missing_value[ivar];
	    handle_error(nc_get_vara_float(ncid, varid[ivar],
					   start, count, var));
	    if (using_nez[ivar]) {
	      handle_error(nc_get_var1_float(ncid, nezid, start, &nez));
	      for (unsigned int igate = 0; igate < ngates; ++igate) {
		if (var[igate] != gap) {
		  var[igate] += (nez_override - nez);
		}
	      }	    
	    }
	    else if (log_scale[ivar]) {
	      for (unsigned int igate = 0; igate < ngates; ++igate) {
		gap = -1.0;
		if (var[igate] <= 0) {
		  var[igate] = gap;
		}
		else {
		  var[igate] = log10(var[igate]) + betacal;
		}
	      }
	    }
	    else if (vfold[ivar] > NO_FLOAT_ATTRIBUTE_TEST) {
	      for (unsigned int igate = 0; igate < ngates; ++igate) {
		float tmpvar = var[igate];
		if (tmpvar != gap && tmpvar < minvalue[ivar]
		    && tmpvar < maxvalue[ivar]) {
		  var[igate] += 2 * vfold[ivar];
		}
		else if (tmpvar != gap && tmpvar > maxvalue[ivar]
			 && tmpvar > minvalue[ivar]) {
		  var[igate] -= 2 * vfold[ivar];
		}
	      }
	    }
	    axes[ivar]->vray(time, nexttime,
			     range, var, ngates, &colormap, 
			     minvalue[ivar], maxvalue[ivar], gap);
	  }
	}
	else {
	  /*** SCANNING DATA ***/
	  int i;
	  float x[4], y[4];
	  float tmpangle = 0;

	  switch (sweep_mode) {
	  case UF_RHI: case UF_COPLANE:
	    tmpangle = elev * PI_OVER_180;
	    if (angle < -999) {
	      dangle = 0.2 * PI_OVER_180; // default:+/-0.2 degs
	      bestangle = tmpangle;
	    }
	    else {
	      dangle = 0.7 * (tmpangle - angle);
	      bestangle = (tmpangle + angle) * 0.5;
	    }
	    angle = tmpangle;
	    break;
	  case UF_PPI:
	    if (angle < -999.0) {
	      dangle = 0.2 * PI_OVER_180; // default:+/-0.2 degs
	    }
	    else {
	    /* Note: the +90.0 is because `angle' is CLOCKWISE from DUE WEST */
	      dangle = 0.7 * ((azim+90.0) * PI_OVER_180 - angle);
	    }
	    angle = (azim+90.0) * PI_OVER_180;
	    break;
	  }

	  for (i = 0; i < ((int) ngates)-1; i++) {
	    float D0 = RADIUS_OF_THE_EARTH, D02;
	    switch (sweep_mode) {
	    case UF_RHI: case UF_COPLANE:
	      /* Correct for the curvature of the earth */
	      D02 = D0 * D0;
	      x[0] = range[i]*cos(bestangle + dangle);
	      y[0] = range[i]*sin(bestangle + dangle)
		+ sqrt(D02 + x[0]*x[0]) - D0;
	      x[1] = range[i+1]*cos(bestangle + dangle);
	      y[1] = range[i+1]*sin(bestangle + dangle)
		+ sqrt(D02 + x[1]*x[1]) - D0;
	      x[2] = range[i+1]*cos(bestangle - dangle);
	      y[2] = range[i+1]*sin(bestangle - dangle)
		+ sqrt(D02 + x[2]*x[2]) - D0;
	      x[3] = range[i]*cos(bestangle - dangle);
	      y[3] = range[i]*sin(bestangle - dangle)
		+ sqrt(D02 + x[3]*x[3]) - D0;
	      break;
	    case UF_PPI:
	      x[0] = -range[i]*cos(angle + dangle);
	      y[0] = range[i]*sin(angle + dangle);
	      x[1] = -range[i+1]*cos(angle + dangle);
	      y[1] = range[i+1]*sin(angle + dangle);
	      x[2] = -range[i+1]*cos(angle - dangle);
	      y[2] = range[i+1]*sin(angle - dangle);
	      x[3] = -range[i]*cos(angle - dangle);
	      y[3] = range[i]*sin(angle - dangle);
	      break;
	    default:
	      x[0] = x[1] = (time-start_time) * 60.0;
	      x[2] = x[3] = (nexttime-start_time) * 60.0;
	      y[0] = y[3] = range[i];
	      y[1] = y[2] = range[i+1];
	    }
	    for (ivar = 0; ivar < max_vars; ivar++) {
	      float value;
	      size_t index[2] = {iray, i};
	      int j;

	      handle_error(nc_get_var1_float(ncid, varid[ivar], index, &value));
	      if (value != missing_value[ivar]) {
		j = (int) (((float) colormap.size())
			   * (value - minvalue[ivar])
			   / (maxvalue[ivar] - minvalue[ivar]));
		if (j < 0) j = 0;
		if (j > colormap.size()-1) j = colormap.size()-1;
		axes[ivar]->polygon(4, x, y, colormap.get(j));
	      }
	    }
	    /* Optimisation for RHIs: skip rest of ray if y[0] out of range */
	    if (sweep_mode != UF_PPI && (x[0] > maxx+1.0 || y[0] > maxy+1.0)) {
	      break;
	    }
	  }
	}
      }

      delete[] var;
      delete[] range; 
      if (system_true) {
	delete[] system_true;
	system_true = NULL;
      }

      dhour = (endhour-starthour > 6) ? floor((endhour-starthour)
					      / 6.0) :
		(1.0+floor((endhour-starthour)/4.0))/2.0;
      for (ivar = 0; ivar < max_vars; ++ivar) {
	axes[ivar]->box();
	if (cloud_data) {
	  axes[ivar]->xticks_time(0, dhour);
	  axes[ivar]->yticks(0,1);
	  axes[ivar]->xlabel("Time (UTC)");
	  axes[ivar]->ylabel("Height (km)");
	}
	else {
	  switch (sweep_mode) {
	  case UF_RHI: case UF_COPLANE:
	    axes[ivar]->xticks(floor(minx*0.1)*10,10);
	    axes[ivar]->yticks(floor(miny*0.5)*2,2);
	    axes[ivar]->xlabel("Distance (km)");
	    axes[ivar]->ylabel("Height (km)");
	    break;
	  case UF_PPI:
	    axes[ivar]->xticks(floor(minx*0.1)*10,10);
	    axes[ivar]->yticks(floor(miny*0.1)*10,10);
	    axes[ivar]->xlabel("Distance east (km)");
	    axes[ivar]->ylabel("Distance north (km)");
	    if (mapfile) {
	      figure[ivar]->clipbox(axes[ivar]->position());
	      MAPPlotCoastline(axes[ivar]);
	      figure[ivar]->clipbox(false);
	    }
	    if (locfile) {
	      figure[ivar]->clipbox(axes[ivar]->position());
	      MAPPlotLocations(axes[ivar]);
	      figure[ivar]->clipbox(false);
	    }
	    break;
	  default:
	    axes[ivar]->xticks(floor(minx), 1);
	    axes[ivar]->yticks(floor(miny*0.5)*2,2);
	    axes[ivar]->xlabel("Approximate time (minutes)");
	    axes[ivar]->ylabel("Range (km)");
	  }
	}
      }
      if (field_title) {
	for (ivar = 0; ivar < max_vars; ++ivar) {
	  if (long_name_override) {
	    long_name = long_name_override;
	  }
	  else {
	    long_name = get_char_attribute(ncid, varid[ivar], "long_name");
	  }
	  figure[ivar]->font = font_loader->load(font_size,
						 string("helvetica"),
						 string("bold"),
						 string("r"),
						 string("adobe"));
	  figure[ivar]->text(x_offset+4, y_offset[ivar]+height-2,
			     long_name, ANCHOR_TOPLEFT);
	  if (long_name && !long_name_override) {
	    delete[] long_name;
	    long_name = NULL;
	  }
	}
      }

      handle_error(nc_close(ncid));

      for (ivar = 0; ivar < max_vars; ++ivar) {
	if (format == FORMAT_PNG && (!one_plot || ivar == 0)) {
	  if (!standard_output) {
	    figure[ivar]->write_png(&ofs[ivar]);
	  }
	  else {
	    figure[ivar]->write_png(&cout);
	  }
	}
	delete axes[ivar];
	if (!one_plot || ivar == 0) {
	  delete figure[ivar];
	}
	if (!standard_output) {
	  ofs[ivar].close();
	}
      }
      if (standard_output) {
	break;
      }
    }
    delete font_loader;
  }
  catch (const char *error_message) {
    cerr << error_message << endl;
    exit(1);
  }
}
