/* addcomments.c -- add comments to a NetCDF file */
#include <stdio.h>
#include <netcdf.h>
#include <string.h>
#include <stdlib.h>

#include "nctools.h"

#define MAX_COMMENTS 100

static int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); nc_close(ncid); exit(-1); \
} 

static
void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.nc [file2.nc ...]\n", argv[0]);
  fprintf(stderr, "Options:\n"
	  "       -global s        append s to global \"" COMMENT_NAME "\" attribute\n"
	  "          -var s1 s2    append s2 to \"" COMMENT_NAME "\" attribute of variable s1\n"
	  "   -global_att s1 s2    append s2 to global attribute s1\n"
	  "      -var_att s1 s2 s3 append s3 to attribute s1 of variable s2\n" 
	  "    -nohistory          do not change the \"history\" global attribute\n"
	  "         -user s        name (and email address) of the user\n");
}

int
main(int argc, char **argv)
{
  char *comments[MAX_COMMENTS];
  char *variables[MAX_COMMENTS];
  char *attributes[MAX_COMMENTS];
  char *user = NULL;
  int ncomments = 0;

  int iarg, ifile;

  char dohistory = 1;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg], "-global") == 0) {
      if (argc <= iarg+1) {
	fprintf(stderr, "Error: global comment not specified\n");
	exit(-1);
      }
      else if (ncomments >= MAX_COMMENTS) {
	fprintf(stderr, "Error: no more than %d comments can be added\n",
		MAX_COMMENTS);
	exit(1);
      }
      else {
	attributes[ncomments] = NULL;
	variables[ncomments] = NULL;
	comments[ncomments++] = argv[++iarg];
      }
    }
    else if (strcmp(argv[iarg], "-var") == 0) {
      if (argc <= iarg+2) {
	fprintf(stderr, "Error: comment and/or variable name not specified\n");
	exit(-1);
      }
      else if (ncomments >= MAX_COMMENTS) {
	fprintf(stderr, "Error: no more than %d comments can be added\n",
		MAX_COMMENTS);
	exit(1);
      }
      else {
	attributes[ncomments] = NULL;
	variables[ncomments] = argv[++iarg];
	comments[ncomments++] = argv[++iarg];
      }
    }
    else if (strcmp(argv[iarg], "-global_att") == 0) {
      if (argc <= iarg+2) {
	fprintf(stderr, "Error: attribute name and/or contents not specified\n");
	exit(-1);
      }
      else if (ncomments >= MAX_COMMENTS) {
	fprintf(stderr, "Error: no more than %d comments can be added\n",
		MAX_COMMENTS);
	exit(1);
      }
      else {
	attributes[ncomments] = argv[++iarg];
	variables[ncomments] = NULL;
	comments[ncomments++] = argv[++iarg];
      }
    }
    else if (strcmp(argv[iarg], "-var_att") == 0) {
      if (argc <= iarg+3) {
	fprintf(stderr, "Error: attribute name, variable name and/or contents not specified\n");
	exit(-1);
      }
      else if (ncomments >= MAX_COMMENTS) {
	fprintf(stderr, "Error: no more than %d comments can be added\n",
		MAX_COMMENTS);
	exit(1);
      }
      else {
	attributes[ncomments] = argv[++iarg];
	variables[ncomments] = argv[++iarg];
	comments[ncomments++] = argv[++iarg];
      }
    }
    else if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  fprintf(stderr, "Warning: user specified more than once:"
		  " `%s' not used\n", user);
	}
	user = argv[++iarg];
      }
      else {
	fprintf(stderr, "Error: user not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg], "-nohistory") == 0) {
      dohistory = 0;
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
    exit(0);
  }

  if (!ncomments) {
    fprintf(stderr, "Warning: no action taken (no comments given)\n");
    exit(0);
  }

  for (ifile = iarg; ifile < argc; ++ifile) {
    int ncid;
    int j;
    char comments_added = 0;

    handle_error(nc_open(argv[ifile], NC_WRITE | NC_NOCLOBBER, &ncid));
    handle_error(nc_redef(ncid));
    
    for (j = 0; j < ncomments; j++) {
      int varid = NC_GLOBAL;
      char skip_var = 0;
      if (variables[j]) {
	if (nc_inq_varid(ncid, variables[j], &varid) != NC_NOERR) {
	  fprintf(stderr, "Warning: %s not found in %s\n",
		  variables[j], argv[ifile]);
	  skip_var = 1;
	}
      }
      if (!skip_var) {
	size_t attlen = 0;
	char *attribute = attributes[j];

	if (!attribute) {
	  /* Existing comments might be in "comments" or "comment" */
	  if (nc_inq_attlen(ncid, varid, "comments", &attlen) == NC_NOERR) {
	    attribute = "comments";
	  }
	  else if (nc_inq_attlen(ncid, varid, "comment", &attlen) == NC_NOERR) {
	    attribute = "comment";
	  }
	  else {
	    /* Use default held in nctools.h */
	    attribute = COMMENT_NAME;
	  }
	}
	nct_append_attribute(ncid, varid, attribute, comments[j], "\n");
	comments_added = 1;
      }
    }
    if (comments_added && dohistory) {
      nct_add_history(ncid, "Comments added", user);
    }

    handle_error(nc_close(ncid));    
  }
  exit(0);
}
